using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Http.Features;
using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Mvc.ModelBinding;
using Microsoft.AspNetCore.WebUtilities;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.Logging;
using Microsoft.Net.Http.Headers;
using System;
using System.Globalization;
using System.IO;
using System.Net;
using System.Text;
using System.Threading.Tasks;

using TestProject.Data;
using TestProject.Filters;
using TestProject.Models;
using TestProject.Utilities;

namespace TestProject.Controllers
{
    /// <summary>
    /// Ʈ Ʈѷ
    /// </summary>
    public class StreamingController : Controller
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// Ʈ  ɼ
        /// </summary>
        /// <remarks>
        /// û  Ϳ  ⺻  ϴ    ֵ ⺻  ɼ ´.
        /// </remarks>
        private static readonly FormOptions _defaultFormOption = new FormOptions();

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// ͺ̽ ؽƮ
        /// </summary>
        private readonly DatabaseContext databaseContext;

        /// <summary>
        ///  ũ 
        /// </summary>
        private readonly long fileSizeLimit;

        /// <summary>
        /// α ϱ
        /// </summary>
        private readonly ILogger<StreamingController> logger;

        /// <summary>
        ///   Ȯ 迭
        /// </summary>
        private readonly string[] allowFileExtensionArray = { ".txt", ".zip", ".jpg" };

        /// <summary>
        /// Ÿ 丮 
        /// </summary>
        private readonly string targetDirectoryPath;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region  - StreamingController(logger, databaseContext, configuration)

        /// <summary>
        /// 
        /// </summary>
        /// <param name="logger">α ϱ</param>
        /// <param name="databaseContext">ͺ̽  ؽƮ</param>
        /// <param name="configuration"></param>
        public StreamingController(ILogger<StreamingController> logger, DatabaseContext databaseContext, IConfiguration configuration)
        {
            this.logger              = logger;
            this.databaseContext     = databaseContext;
            this.fileSizeLimit       = configuration.GetValue<long>("FileSizeLimit");
            this.targetDirectoryPath = configuration.GetValue<string>("UploadDirectoryPath");
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        //  ε ޼ҵ :
        //
        // 1.  ū  ó Ϸ    ε ȰȭѴ.
        //
        // 2. Ϲ   ū û  ۵ȴ.
        //    û   а  ʱ  ū   ۵ȴ.
        //      ū ʹ  û  ū ã   б ư.

        #region ͺ̽ εϱ - UploadDatabase()

        /// <summary>
        /// ͺ̽ εϱ
        /// </summary>
        /// <returns>׼  ½ũ</returns>
        [HttpPost]
        [DisableFormValueModelBinding]
        [ValidateAntiForgeryToken]
        public async Task<IActionResult> UploadDatabase()
        {
            if(!MultipartRequestHelper.IsMultipartContentType(Request.ContentType))
            {
                ModelState.AddModelError("File", $"The request couldn't be processed (Error 1).");

                return BadRequest(ModelState);
            }

            // û(formAccumulator)   Ű-  Ѵ.
            KeyValueAccumulator formAccumulator              = new KeyValueAccumulator();
            string              sourceFileName               = string.Empty;
            string              targetFileName               = string.Empty;
            byte[]              streamedFileContentByteArray = new byte[0];

            string boundary = MultipartRequestHelper.GetBoundary
            (
                MediaTypeHeaderValue.Parse(Request.ContentType),
                _defaultFormOption.MultipartBoundaryLengthLimit
            );

            MultipartReader multipartReader = new MultipartReader(boundary, HttpContext.Request.Body);

            MultipartSection multipartSection = await multipartReader.ReadNextSectionAsync();

            while(multipartSection != null)
            {
                bool hasContentDispositionHeaderValue = ContentDispositionHeaderValue.TryParse
                (
                    multipartSection.ContentDisposition,
                    out var contentDispositionHeaderValue
                );

                if(hasContentDispositionHeaderValue)
                {
                    if(MultipartRequestHelper.HasFileContentDisposition(contentDispositionHeaderValue))
                    {
                        targetFileName = contentDispositionHeaderValue.FileName.Value;

                        // Ŭ̾Ʈ   ̸ ŷ ʴ´.
                        //  ̸ ǥϷ  HTML ڵѴ.
                        sourceFileName = WebUtility.HtmlEncode(contentDispositionHeaderValue.FileName.Value);

                        streamedFileContentByteArray = await FileHelper.ProcessStreamedFile
                        (
                            multipartSection,
                            contentDispositionHeaderValue,
                            ModelState,
                            this.allowFileExtensionArray,
                            this.fileSizeLimit
                        );

                        if(!ModelState.IsValid)
                        {
                            return BadRequest(ModelState);
                        }
                    }
                    else if(MultipartRequestHelper.HasFormDataContentDisposition(contentDispositionHeaderValue))
                    {
                        //  κ    ̹ ǾǷ Ű ̸ ̸  ʴ´.
                        string   key      = HeaderUtilities.RemoveQuotes(contentDispositionHeaderValue.Name).Value;
                        Encoding encoding = GetEncoding(multipartSection);

                        if(encoding == null)
                        {
                            ModelState.AddModelError("File", $"The request couldn't be processed (Error 2).");

                            return BadRequest(ModelState);
                        }

                        using
                        (
                            StreamReader streamReader = new StreamReader
                            (
                                multipartSection.Body,
                                encoding,
                                detectEncodingFromByteOrderMarks : true,
                                bufferSize : 1024,
                                leaveOpen : true
                            )
                        )
                        {
                            //    MultipartBodyLengthLimit  ȴ.
                            string value = await streamReader.ReadToEndAsync();

                            if(string.Equals(value, "undefined", StringComparison.OrdinalIgnoreCase))
                            {
                                value = string.Empty;
                            }

                            formAccumulator.Append(key, value);

                            if(formAccumulator.ValueCount > _defaultFormOption.ValueCountLimit)
                            {
                                // _defaultFormOption.ValueCountLimit  Ű   ʰǾϴ.
                                ModelState.AddModelError("File", $"The request couldn't be processed (Error 3).");

                                return BadRequest(ModelState);
                            }
                        }
                    }
                }

                // Һ         д´.
                multipartSection = await multipartReader.ReadNextSectionAsync();
            }

            //  ͸ 𵨿 ε Ѵ.
            FormData formData = new FormData();

            FormValueProvider formValueProvider = new FormValueProvider
            (
                BindingSource.Form,
                new FormCollection(formAccumulator.GetResults()),
                CultureInfo.CurrentCulture
            );

            bool isBindingSuccessful = await TryUpdateModelAsync(formData, prefix : "", valueProvider : formValueProvider);

            if(!isBindingSuccessful)
            {
                ModelState.AddModelError("File", "The request couldn't be processed (Error 5).");

                return BadRequest(ModelState);
            }

            // !
            //     ĵ ʰ  ȴ.
            // κ δ ó  ٿεϰų ٸ ýۿ   ֵϱ 
            // ̷ /ֿ  ĳ API Ͽ ȴ.

            FileItemModel fileItem = new FileItemModel()
            {
                Content    = streamedFileContentByteArray,
                FileName   = targetFileName,
                Note       = formData.Note,
                Size       = streamedFileContentByteArray.Length, 
                UploadTime = DateTime.UtcNow
            };

            this.databaseContext.FileItemDBSet.Add(fileItem);

            await this.databaseContext.SaveChangesAsync();

            return Created(nameof(StreamingController), null);
        }

        #endregion
        #region   εϱ - UploadPhysical()

        /// <summary>
        ///   εϱ
        /// </summary>
        /// <returns>׼  ½ũ</returns>
        [HttpPost]
        [DisableFormValueModelBinding]
        [ValidateAntiForgeryToken]
        public async Task<IActionResult> UploadPhysical()
        {
            if(!MultipartRequestHelper.IsMultipartContentType(Request.ContentType))
            {
                ModelState.AddModelError("File", $"The request couldn't be processed (Error 1).");

                return BadRequest(ModelState);
            }

            string boundary = MultipartRequestHelper.GetBoundary
            (
                MediaTypeHeaderValue.Parse(Request.ContentType),
                _defaultFormOption.MultipartBoundaryLengthLimit
            );

            MultipartReader multipartReader = new MultipartReader(boundary, HttpContext.Request.Body);

            MultipartSection multipartSection = await multipartReader.ReadNextSectionAsync();

            while(multipartSection != null)
            {
                bool hasContentDispositionHeaderValue = ContentDispositionHeaderValue.TryParse
                (
                    multipartSection.ContentDisposition,
                    out var contentDispositionHeaderValue
                );

                if(hasContentDispositionHeaderValue)
                {
                    if(!MultipartRequestHelper.HasFileContentDisposition(contentDispositionHeaderValue))
                    {
                        ModelState.AddModelError("File", $"The request couldn't be processed (Error 2).");

                        return BadRequest(ModelState);
                    }
                    else
                    {
                        string sourceFileName = WebUtility.HtmlEncode(contentDispositionHeaderValue.FileName.Value);
                        string targetFileName = sourceFileName;

                        // !
                        //     ĵ ʰ  ȴ.
                        // κ δ ó  ٿεϰų ٸ ýۿ   ֵ ϱ 
                        // ̷ /ֿ  ĳ API Ͽ ȴ.

                        byte[] streamedFileContentByteArray = await FileHelper.ProcessStreamedFile
                        (
                            multipartSection,
                            contentDispositionHeaderValue,
                            ModelState, 
                            this.allowFileExtensionArray,
                            this.fileSizeLimit
                        );

                        if(!ModelState.IsValid)
                        {
                            return BadRequest(ModelState);
                        }

                        using(FileStream targetStream = System.IO.File.Create(Path.Combine(this.targetDirectoryPath, targetFileName)))
                        {
                            await targetStream.WriteAsync(streamedFileContentByteArray);

                            this.logger.LogInformation($"Uploaded file '{sourceFileName}' saved to '{this.targetDirectoryPath}' as {targetFileName}"); 
                        }
                    }
                }

                // Һ         д´.
                multipartSection = await multipartReader.ReadNextSectionAsync();
            }

            return Created(nameof(StreamingController), null);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region ڵ ϱ - GetEncoding(multipartSection)

        /// <summary>
        /// ڵ ϱ
        /// </summary>
        /// <param name="multipartSection">Ƽ Ʈ </param>
        /// <returns>ڵ</returns>
        private static Encoding GetEncoding(MultipartSection multipartSection)
        {
            var hasMediaTypeHeaderValue = MediaTypeHeaderValue.TryParse(multipartSection.ContentType, out var mediaTypeHeaderValue);

            // UTF-7   ؼ ȵȴ.
            // UTF-8 κ  Ѵ.
            if(!hasMediaTypeHeaderValue || Encoding.UTF7.Equals(mediaTypeHeaderValue.Encoding))
            {
                return Encoding.UTF8;
            }

            return mediaTypeHeaderValue.Encoding;
        }

        #endregion
    }
}