﻿using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Http.Features;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.FileProviders;
using Microsoft.Extensions.Hosting;

using TestProject.Data;
using TestProject.Filters;

namespace TestProject
{
    /// <summary>
    /// 시작
    /// </summary>
    public class Startup
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 구성 - Configuration

        /// <summary>
        /// 구성
        /// </summary>
        public IConfiguration Configuration { get; }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - Startup(configuration)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="configuration">구성</param>
        public Startup(IConfiguration configuration)
        {
            Configuration = configuration;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 서비스 컬렉션 구성하기 - ConfigureServices(services)

        /// <summary>
        /// 서비스 컬렉션 구성하기
        /// </summary>
        /// <param name="services">서비스 컬렉션</param>
        public void ConfigureServices(IServiceCollection services)
        {
            long fileSizeLimit = Configuration.GetValue<long>("FileSizeLimit");

            services.AddControllers();

            services.AddRazorPages
            (
                options =>
                {
                    options.Conventions.AddPageApplicationModelConvention
                    (
                        "/UploadStreamOneFileToDatabase",
                        model =>
                        {
                            model.Filters.Add(new GenerateAntiforgeryTokenCookieAttribute());
                            model.Filters.Add(new DisableFormValueModelBindingAttribute());
                        }
                    );

                    options.Conventions.AddPageApplicationModelConvention
                    (
                        "/UploadStreamOneFileToPhysicalStorage",
                        model =>
                        {
                            model.Filters.Add(new GenerateAntiforgeryTokenCookieAttribute());
                            model.Filters.Add(new DisableFormValueModelBindingAttribute());
                        }
                    );
                }
            );

            // Kestrel 사용시
            services.Configure<FormOptions>(options => { options.MultipartBodyLengthLimit = fileSizeLimit; });

            // IIS Express 사용시
            //services.Configure<IISServerOptions>(options => { options.MaxRequestBodySize = fileSizeLimit; });

            PhysicalFileProvider physicalFileProvider = new PhysicalFileProvider(Configuration.GetValue<string>("UploadDirectoryPath"));

            services.AddSingleton<IFileProvider>(physicalFileProvider);

            services.AddDbContext<DatabaseContext>(options => options.UseInMemoryDatabase("InMemoryDb"));
        }

        #endregion
        #region 구성하기 - Configure(app, environment)

        /// <summary>
        /// 구성하기
        /// </summary>
        /// <param name="app">애플리케이션 빌더</param>
        /// <param name="environment">웹 호스트 환경</param>
        public void Configure(IApplicationBuilder app, IWebHostEnvironment environment)
        {
            if(environment.IsDevelopment())
            {
                app.UseDeveloperExceptionPage();
            }
            else
            {
                app.UseExceptionHandler("/Error");
            }

            app.UseStaticFiles();

            app.UseRouting();

            app.UseEndpoints
            (
                endpoints =>
                {
                    endpoints.MapControllerRoute("default", "{controller=Home}/{action=Index}/{id?}");

                    endpoints.MapRazorPages();
                }
            );
        }

        #endregion
    }
}