using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc.ModelBinding;
using Microsoft.AspNetCore.WebUtilities;
using Microsoft.Net.Http.Headers;
using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.IO;
using System.Linq;
using System.Net;
using System.Reflection;
using System.Threading.Tasks;

namespace TestProject.Utilities
{
    /// <summary>
    ///  
    /// </summary>
    public static class FileHelper
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        ///   Ʈ 迭
        /// </summary>
        /// <remarks>
        /// IsValidFileExtensionAndSignature ޼ Ư ڸ Ȯؾϴ 
        /// _allowedCharacterByteArray ʵ忡 ڸ մϴ.
        /// </remarks>
        private static readonly byte[] _allowCharacterByteArray = { };

        /// <summary>
        ///  ñ״ó ųʸ
        /// </summary>
        /// <remarks>
        ///       ͺ̽ (https://www.filesignatures.net/)  ߰Ϸ  Ŀ    Ѵ.
        /// </remarks>
        private static readonly Dictionary<string, List<byte[]>> _fileSignatureDictionary = new Dictionary<string, List<byte[]>>
        {
            {
                ".gif",
                new List<byte[]>
                {
                    new byte[] { 0x47, 0x49, 0x46, 0x38 }
                }
            },
            {
                ".png",
                new List<byte[]>
                {
                    new byte[] { 0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A }
                }
            },
            {
                ".jpeg",
                new List<byte[]>
                {
                    new byte[] { 0xFF, 0xD8, 0xFF, 0xE0 },
                    new byte[] { 0xFF, 0xD8, 0xFF, 0xE2 },
                    new byte[] { 0xFF, 0xD8, 0xFF, 0xE3 }
                }
            },
            {
                ".jpg",
                new List<byte[]>
                {
                    new byte[] { 0xFF, 0xD8, 0xFF, 0xE0 },
                    new byte[] { 0xFF, 0xD8, 0xFF, 0xE1 },
                    new byte[] { 0xFF, 0xD8, 0xFF, 0xE8 }
                }
            },
            {
                ".zip",
                new List<byte[]> 
                {
                    new byte[] { 0x50, 0x4B, 0x03, 0x04             }, 
                    new byte[] { 0x50, 0x4B, 0x4C, 0x49, 0x54, 0x45 },
                    new byte[] { 0x50, 0x4B, 0x53, 0x70, 0x58       },
                    new byte[] { 0x50, 0x4B, 0x05, 0x06             },
                    new byte[] { 0x50, 0x4B, 0x07, 0x08             },
                    new byte[] { 0x57, 0x69, 0x6E, 0x5A, 0x69, 0x70 }
                }
            }
        };

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        // !
        //   ó    ˻ ʴ´.
        // κ δ ó ̷ /ֿ  ĳ API
        // ڳ ٸ ýۿ    ֵ ϱ  Ͽ ȴ.
        // ڼ    ۰ Բ Ǵ  Ѵ.
        #region Ͽ óϱ - ProcessFormFile<T>(formFile, modelStateDictionary, sourceFileExtensionArray, fileSizeLimit)

        /// <summary>
        /// Ͽ óϱ
        /// </summary>
        /// <typeparam name="T">Ÿ</typeparam>
        /// <param name="formFile"> </param>
        /// <param name="modelStateDictionary">  ųʸ</param>
        /// <param name="sourceFileExtensionArray">ҽ  Ȯ 迭</param>
        /// <param name="fileSizeLimit"> ũ </param>
        /// <returns>Ʈ 迭 ½ũ</returns>
        public static async Task<byte[]> ProcessFormFile<T>
        (
            IFormFile            formFile,
            ModelStateDictionary modelStateDictionary,
            string[]             sourceFileExtensionArray,
            long                 fileSizeLimit
        )
        {
            string fieldDisplayName = string.Empty;

            // ÷ Ͽ  IFormFile   Ӽ ǥ ̸ ´.
            // ǥ ̸ ã     ޽ ǥ ̸ ǥõ ʴ´.
            MemberInfo propertyMemberInfo = typeof(T).GetProperty
            (
                formFile.Name.Substring
                (
                    formFile.Name.IndexOf(".", StringComparison.Ordinal) + 1
                )
            );

            if(propertyMemberInfo != null)
            {
                if(propertyMemberInfo.GetCustomAttribute(typeof(DisplayAttribute)) is DisplayAttribute displayAttribute)
                {
                    fieldDisplayName = $"{displayAttribute.Name} ";
                }
            }

            // Ŭ̾Ʈ   ̸ ŷ ʴ´.
            //  ̸ ǥϷ  HTML ڵѴ.
            string trustedFileNameForDisplay = WebUtility.HtmlEncode(formFile.FileName);

            //  ̸ ȮѴ.
            //  ˻  BOM ִ   ʴ´.
            if(formFile.Length == 0)
            {
                modelStateDictionary.AddModelError
                (
                    formFile.Name,
                    $"{fieldDisplayName}({trustedFileNameForDisplay}) is empty."
                );

                return new byte[0];
            }
            
            if(formFile.Length > fileSizeLimit)
            {
                long fileSizeLimitMB = fileSizeLimit / 1048576;

                modelStateDictionary.AddModelError
                (
                    formFile.Name,
                    $"{fieldDisplayName}({trustedFileNameForDisplay}) exceeds {fileSizeLimitMB:N1} MB."
                );

                return new byte[0];
            }

            try
            {
                using(MemoryStream memoryStream = new MemoryStream())
                {
                    await formFile.CopyToAsync(memoryStream);

                    //    BOM̾
                    // BOM     ִ   ̸ ȮѴ.
                    if(memoryStream.Length == 0)
                    {
                        modelStateDictionary.AddModelError
                        (
                            formFile.Name,
                            $"{fieldDisplayName}({trustedFileNameForDisplay}) is empty."
                        );
                    }

                    if(!IsValidFileExtensionAndSignature(formFile.FileName, memoryStream, sourceFileExtensionArray))
                    {
                        modelStateDictionary.AddModelError
                        (
                            formFile.Name,
                            $"{fieldDisplayName}({trustedFileNameForDisplay}) file type isn't permitted or " +
                                "the file's signature doesn't match the file's extension."
                        );
                    }
                    else
                    {
                        return memoryStream.ToArray();
                    }
                }
            }
            catch(Exception exception)
            {
                modelStateDictionary.AddModelError
                (
                    formFile.Name,
                    $"{fieldDisplayName}({trustedFileNameForDisplay}) upload failed. " +
                        "Please contact the Help Desk for support. Error : {exception.HResult}"
                );
            }

            return new byte[0];
        }

        #endregion
        #region Ʈ  óϱ - ProcessStreamedFile(multipartSection, contentDispositionHeaderValue, modelStateDictionary, sourceFileExtensionArray, fileSizeLimit)

        /// <summary>
        /// Ʈ  óϱ
        /// </summary>
        /// <param name="multipartSection">Ƽ Ʈ </param>
        /// <param name="contentDispositionHeaderValue">Ʈ ġ  </param>
        /// <param name="modelStateDictionary">  ųʸ</param>
        /// <param name="sourceFileExtensionArray">ҽ  Ȯ 迭</param>
        /// <param name="fileSizeLimit"> ũ </param>
        /// <returns>Ʈ 迭 ½ũ</returns>
        public static async Task<byte[]> ProcessStreamedFile
        (
            MultipartSection              multipartSection,
            ContentDispositionHeaderValue contentDispositionHeaderValue,
            ModelStateDictionary          modelStateDictionary,
            string[]                      sourceFileExtensionArray,
            long                          fileSizeLimit
        )
        {
            try
            {
                using(MemoryStream memoryStream = new MemoryStream())
                {
                    await multipartSection.Body.CopyToAsync(memoryStream);

                    //   ְų ũ  ʰϴ ȮѴ.
                    if(memoryStream.Length == 0)
                    {
                        modelStateDictionary.AddModelError("File", "The file is empty.");
                    }
                    else if(memoryStream.Length > fileSizeLimit)
                    {
                        long fileSizeLimitMB = fileSizeLimit / 1048576;

                        modelStateDictionary.AddModelError("File", $"The file exceeds {fileSizeLimitMB:N1} MB.");
                    }
                    else if
                    (
                        !IsValidFileExtensionAndSignature
                        (
                            contentDispositionHeaderValue.FileName.Value,
                            memoryStream,
                            sourceFileExtensionArray
                        )
                    )
                    {
                        modelStateDictionary.AddModelError
                        (
                            "File",
                            "The file type isn't permitted or the file's signature doesn't match the file's extension."
                        );
                    }
                    else
                    {
                        return memoryStream.ToArray();
                    }
                }
            }
            catch(Exception exception)
            {
                modelStateDictionary.AddModelError
                (
                    "File",
                    $"The upload failed. Please contact the Help Desk for support. Error: {exception.HResult}"
                );
            }

            return new byte[0];
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region  Ȯ/ñ״ó ȿ  ϱ - IsValidFileExtensionAndSignature(fileName, stream, sourceFileExtensionArray)

        /// <summary>
        ///  Ȯ/ñ״ó ȿ  ϱ
        /// </summary>
        /// <param name="fileName">ϸ</param>
        /// <param name="stream">Ʈ</param>
        /// <param name="sourceFileExtensionArray">ҽ  Ȯ 迭</param>
        /// <returns> Ȯ/ñ״ó ȿ </returns>
        private static bool IsValidFileExtensionAndSignature(string fileName, Stream stream, string[] sourceFileExtensionArray)
        {
            if(string.IsNullOrEmpty(fileName) || stream == null || stream.Length == 0)
            {
                return false;
            }

            string fileExtension = Path.GetExtension(fileName).ToLowerInvariant();

            if(string.IsNullOrEmpty(fileExtension) || !sourceFileExtensionArray.Contains(fileExtension))
            {
                return false;
            }

            stream.Position = 0;

            using(BinaryReader reader = new BinaryReader(stream))
            {
                if(fileExtension.Equals(".txt") || fileExtension.Equals(".csv") || fileExtension.Equals(".prn"))
                {
                    if(_allowCharacterByteArray.Length == 0)
                    {
                        // ڸ ASCII ڵ Ѵ.
                        for(long i = 0L; i < stream.Length; i++)
                        {
                            if(reader.ReadByte() > sbyte.MaxValue)
                            {
                                return false;
                            }
                        }
                    }
                    else
                    {
                        // ڸ ASCII ڵ  _allowedCharacterByteArray 迭  մϴ.
                        for(long i = 0L; i < stream.Length; i++)
                        {
                            byte byteValue = reader.ReadByte();

                            if(byteValue > sbyte.MaxValue || !_allowCharacterByteArray.Contains(byteValue))
                            {
                                return false;
                            }
                        }
                    }

                    return true;
                }

                // _fileSignatureDictionary     ؾ ϴ 
                //  ڵ  ּ ó Ѵ.
                // ýۿ Ϸ   Ŀ   
                // Ͽ    ߰ϰ   ˻縦 ϴ  .
                //if(!_fileSignatureDictionary.ContainsKey(fileExtension))
                //{
                //    return true;
                //}
                

                //   Ȯ
                // --------------
                // _fileSignatureDictionary    Ͽ
                //  ڵ Է    ׽ƮѴ.
                List<byte[]> signatureList = _fileSignatureDictionary[fileExtension];

                byte[] headerByteArray = reader.ReadBytes(signatureList.Max(m => m.Length));

                return signatureList.Any(signature => headerByteArray.Take(signature.Length).SequenceEqual(signature));
            }
        }

        #endregion
    }
}