using Microsoft.Net.Http.Headers;
using System;
using System.IO;

namespace TestProject.Utilities
{
    /// <summary>
    /// Ƽ Ʈ û 
    /// </summary>
    public static class MultipartRequestHelper
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        // Content-Type : multipart/form-data; boundary="----WebKitFormBoundarymx2fSWqWSd0OxQqq"
        // https://tools.ietf.org/html/rfc2046#section-5.1 翡 70ڰ ո ̶  ִ.
        #region  ϱ - GetBoundary(mediaTypeHeaderValue, boundaryLengthLimit)

        /// <summary>
        ///  ϱ
        /// </summary>
        /// <param name="mediaTypeHeaderValue">̵ Ÿ  </param>
        /// <param name="boundaryLengthLimit">  </param>
        /// <returns></returns>
        public static string GetBoundary(MediaTypeHeaderValue mediaTypeHeaderValue, int boundaryLengthLimit)
        {
            string boundary = HeaderUtilities.RemoveQuotes(mediaTypeHeaderValue.Boundary).Value;

            if(string.IsNullOrWhiteSpace(boundary))
            {
                throw new InvalidDataException("Missing content-type boundary.");
            }

            if(boundary.Length > boundaryLengthLimit)
            {
                throw new InvalidDataException($"Multipart boundary length limit {boundaryLengthLimit} exceeded.");
            }

            return boundary;
        }

        #endregion
        #region Ƽ Ʈ Ʈ Ÿ  ϱ - IsMultipartContentType(contentType)

        /// <summary>
        /// Ƽ Ʈ Ʈ Ÿ  ϱ
        /// </summary>
        /// <param name="contentType">Ʈ Ÿ</param>
        /// <returns>Ƽ Ʈ Ʈ Ÿ </returns>
        public static bool IsMultipartContentType(string contentType)
        {
            return !string.IsNullOrEmpty(contentType) && contentType.IndexOf("multipart/", StringComparison.OrdinalIgnoreCase) >= 0;
        }

        #endregion
        #region   Ʈ ġ   ϱ - HasFormDataContentDisposition(contentDispositionHeaderValue)

        /// <summary>
        ///   Ʈ ġ   ϱ
        /// </summary>
        /// <param name="contentDispositionHeaderValue">Ʈ ġ  </param>
        /// <returns>  Ʈ ġ  </returns>
        public static bool HasFormDataContentDisposition(ContentDispositionHeaderValue contentDispositionHeaderValue)
        {
            // Content-Disposition: form-data; name="key";
            return contentDispositionHeaderValue != null                              &&
                   contentDispositionHeaderValue.DispositionType.Equals("form-data")  &&
                   string.IsNullOrEmpty(contentDispositionHeaderValue.FileName.Value) &&
                   string.IsNullOrEmpty(contentDispositionHeaderValue.FileNameStar.Value);
        }

        #endregion
        #region  Ʈ ġ   ϱ - HasFileContentDisposition(contentDispositionHeaderValue)

        /// <summary>
        ///  Ʈ ġ   ϱ
        /// </summary>
        /// <param name="contentDispositionHeaderValue"></param>
        /// <returns> Ʈ ġ  </returns>
        public static bool HasFileContentDisposition(ContentDispositionHeaderValue contentDispositionHeaderValue)
        {
            // Content-Disposition: form-data; name="file1"; filename="sample.jpg"
            return contentDispositionHeaderValue != null                             &&
                   contentDispositionHeaderValue.DispositionType.Equals("form-data") &&
                   (!string.IsNullOrEmpty(contentDispositionHeaderValue.FileName.Value) || !string.IsNullOrEmpty(contentDispositionHeaderValue.FileNameStar.Value));
        }

        #endregion
    }
}